"""Re-usable utility functions used by this package."""

from __future__ import annotations

import os

from dataclasses import dataclass, field
from importlib import resources as impl_resources
from typing import TYPE_CHECKING

import yaml

from ansible_creator.constants import GLOBAL_TEMPLATE_VARS


if TYPE_CHECKING:
    from ansible_creator.compat import Traversable
    from ansible_creator.output import Output
    from ansible_creator.templar import Templar

PATH_REPLACERS = {
    "project_org": "scm_org",
    "project_repo": "scm_project",
}


@dataclass
class TermFeatures:
    """Terminal features."""

    color: bool
    links: bool

    def any_enabled(self: TermFeatures) -> bool:
        """Return True if any features are enabled."""
        return any((self.color, self.links))


def expand_path(path: str) -> str:
    """Resolve absolute path.

    :param path: Path to expand.
    :returns: Expanded absolute path.
    """
    return os.path.abspath(
        os.path.expanduser(os.path.expandvars(path)),
    )


@dataclass
class Copier:
    """Configuration for the Copier class."""

    resources: list[str]
    """ list of resource containers to copy"""
    resource_id: str
    """the id of the resource to copy"""
    dest: str
    """the destination path to copy resources to"""
    output: Output
    """an instance of the Output class"""
    allow_overwrite: list[str] | None = None
    """a list of paths that should be overwritten at destination"""
    index: int = 0
    """index of the current resource being copied"""
    resource_root: str = "ansible_creator.resources"
    """root path for the resources"""
    templar: Templar | None = None
    """an instance of the Templar class"""
    template_data: dict[str, str] = field(default_factory=dict)
    """a dictionary containing the original data to render templates with"""

    @property
    def resource(self: Copier) -> str:
        """Return the current resource being copied."""
        return self.resources[self.index]

    def _recursive_copy(
        self: Copier,
        root: Traversable,
        template_data: dict[str, str],
    ) -> None:
        """Recursively traverses a resource container and copies content to destination.

        :param root: A traversable object representing root of the container to copy.
        :param copier_config: Configuration for the Copier class.
        :param template_data: A dictionary containing current data to render templates with.
        """
        self.output.debug(msg=f"current root set to {root}")

        for obj in root.iterdir():
            overwrite = False
            # resource names may have a . but directories use / in the path
            dest_name = str(obj).split(
                self.resource.replace(".", "/") + "/",
                maxsplit=1,
            )[-1]
            dest_path = os.path.join(self.dest, dest_name)
            if (self.allow_overwrite) and (dest_name in self.allow_overwrite):
                overwrite = True
            # replace placeholders in destination path with real values
            for key, val in PATH_REPLACERS.items():
                if key in dest_path and template_data:
                    dest_path = dest_path.replace(key, template_data.get(val, ""))

            if obj.is_dir():
                if not os.path.exists(dest_path):
                    os.makedirs(dest_path)

                # recursively copy the directory
                self._recursive_copy(
                    root=obj,
                    template_data=template_data,
                )

            elif obj.is_file():
                if obj.name == "__meta__.yml":
                    continue
                # remove .j2 suffix at destination
                dest_file = os.path.join(
                    self.dest,
                    dest_path.split(".j2", maxsplit=1)[0],
                )
                self.output.debug(msg=f"dest file is {dest_file}")

                # write at destination only if missing or belongs to overwrite list
                if not os.path.exists(dest_file) or overwrite:
                    content = obj.read_text(encoding="utf-8")
                    # only render as templates if both of these are provided
                    # templating is not mandatory
                    if self.templar and template_data:
                        content = self.templar.render_from_content(
                            template=content,
                            data=template_data,
                        )
                    with open(dest_file, "w", encoding="utf-8") as df_handle:
                        df_handle.write(content)

    def _per_container(self: Copier) -> None:
        """Copy files and directories from a possibly nested source to a destination.

        :param copier_config: Configuration for the Copier class.

        :raises CreatorError: if allow_overwrite is not a list.
        """
        self.output.debug(
            msg=f"starting recursive copy with source container '{self.resource}'",
        )
        self.output.debug(msg=f"allow_overwrite set to {self.allow_overwrite}")

        # Include the global template variables
        self.template_data.update(GLOBAL_TEMPLATE_VARS)

        # Copy the template data to not pollute the original
        template_data = self.template_data.copy()

        # Collect and template any resource specific variables
        meta_file = (
            impl_resources.files(f"{self.resource_root}.{self.resource}")
            / "__meta__.yml"
        )
        try:
            with meta_file.open("r", encoding="utf-8") as meta_fileh:
                self.output.debug(
                    msg=f"loading resource specific vars from {meta_file}",
                )
                meta = yaml.safe_load(meta_fileh.read())
        except FileNotFoundError:
            meta = {}
            self.output.debug(msg="no resource specific vars found")

        found = meta.get(self.resource_id, {})
        for key, value in found.items():
            if value["template"] and self.templar:
                serialized = yaml.dump(value["value"])
                templated = self.templar.render_from_content(
                    template=serialized,
                    data=template_data,
                )
                deserialized = yaml.safe_load(templated)
                template_data.update({key: deserialized})
            else:
                template_data.update({key: value["value"]})

        self._recursive_copy(
            root=impl_resources.files(f"{self.resource_root}.{self.resource}"),
            template_data=template_data,
        )

    def copy_containers(
        self: Copier,
    ) -> None:
        """Copy multiple containers to destination.

        :param copier_config: Configuration for the Copier class.
        """
        for i in range(len(self.resources)):
            self.index = i
            self._per_container()



"""Definitions for ansible-creator init action."""

from __future__ import annotations

import os
import shutil

from typing import TYPE_CHECKING

from ansible_creator.exceptions import CreatorError
from ansible_creator.templar import Templar
from ansible_creator.utils import Copier


if TYPE_CHECKING:
    from ansible_creator.config import Config
    from ansible_creator.output import Output


class Init:
    """Class representing ansible-creator init subcommand."""

    def __init__(
        self: Init,
        config: Config,
    ) -> None:
        """Initialize the init action.

        :param config: App configuration object.
        :param output: Output class object.
        """
        self._namespace: str = config.namespace
        self._collection_name: str = config.collection_name
        self._init_path: str = config.init_path
        self._force = config.force
        self._creator_version = config.creator_version
        self._project = config.project
        self._scm_org = config.scm_org
        self._scm_project = config.scm_project
        self._templar = Templar()
        self.output: Output = config.output

    def run(self: Init) -> None:
        """Start scaffolding skeleton.

        :raises CreatorError: if computed collection path is an existing directory or file.
        """
        if self._init_path.endswith("collections/ansible_collections"):
            self._init_path = os.path.join(
                self._init_path,
                self._namespace,
                self._collection_name,
            )

        self.output.debug(msg=f"final collection path set to {self._init_path}")

        # check if init_path already exists
        if os.path.exists(self._init_path):
            # init-path exists and is a file
            if os.path.isfile(self._init_path):
                msg = f"the path {self._init_path} already exists, but is a file - aborting"
                raise CreatorError(
                    msg,
                )
            if os.listdir(self._init_path):
                # init-path exists and is not empty, but user did not request --force
                if not self._force:
                    msg = (
                        f"The directory {self._init_path} is not empty.\n"
                        f"You can use --force to re-initialize this directory."
                        f"\nHowever it will delete ALL existing contents in it."
                    )
                    raise CreatorError(msg)

                # user requested --force, re-initializing existing directory
                self.output.warning(
                    f"re-initializing existing directory {self._init_path}",
                )
                for root, dirs, files in os.walk(self._init_path, topdown=True):
                    for old_dir in dirs:
                        path = os.path.join(root, old_dir)
                        self.output.debug(f"removing tree {old_dir}")
                        shutil.rmtree(path)
                    for old_file in files:
                        path = os.path.join(root, old_file)
                        self.output.debug(f"removing file {old_file}")
                        os.unlink(path)

        # if init_path does not exist, create it
        if not os.path.exists(self._init_path):
            self.output.debug(msg=f"creating new directory at {self._init_path}")
            os.makedirs(self._init_path)

        if self._project == "collection":
            # copy new_collection container to destination, templating files when found
            self.output.debug(msg="started copying collection skeleton to destination")
            copier = Copier(
                resources=[
                    "new_collection",
                    "common.devcontainer",
                    "common.devfile",
                    "common.gitignore",
                ],
                resource_id="new_collection",
                dest=self._init_path,
                output=self.output,
                templar=self._templar,
                template_data={
                    "namespace": self._namespace,
                    "collection_name": self._collection_name,
                    "creator_version": self._creator_version,
                },
            )
            copier.copy_containers()

            self.output.note(
                f"collection {self._namespace}.{self._collection_name} "
                f"created at {self._init_path}",
            )

        else:
            self.output.debug(
                msg="started copying ansible-project skeleton to destination",
            )
            copier = Copier(
                resources=[
                    "ansible_project",
                    "common.devcontainer",
                    "common.devfile",
                    "common.gitignore",
                ],
                resource_id="ansible_project",
                dest=self._init_path,
                output=self.output,
                templar=self._templar,
                template_data={
                    "scm_org": self._scm_org,
                    "scm_project": self._scm_project,
                    "creator_version": self._creator_version,
                },
            )
            copier.copy_containers()

            self.output.note(
                f"ansible project created at {self._init_path}",
            )



"""Application configuration class for ansible-creator."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ansible_creator.exceptions import CreatorError
from ansible_creator.utils import expand_path


if TYPE_CHECKING:
    from ansible_creator.output import Output


@dataclass(frozen=True)
class Config:
    """The application configuration for ansible-creator."""

    creator_version: str
    output: Output
    subcommand: str

    collection: str = ""
    force: bool = False
    init_path: str = "./"
    project: str = ""
    scm_org: str = ""
    scm_project: str = ""

    # TO-DO: Add instance variables for other 'create' and 'sample'

    collection_name: str = ""
    namespace: str = ""

    def __post_init__(self: Config) -> None:
        """Post process config values."""
        # Validation for: ansible-creator init
        if not self.collection and self.project == "collection":
            msg = "The argument 'collection' is required when scaffolding a collection."
            raise CreatorError(msg)

        # Validation for: ansible-creator init --project=ansible-project
        if self.project == "ansible-project" and (
            self.scm_org is None or self.scm_project is None
        ):
            msg = (
                "Parameters 'scm-org' and 'scm-project' are required when "
                "scaffolding an ansible-project."
            )
            raise CreatorError(msg)

        # Validation for: ansible-creator init testorg.testname --scm-org=weather
        # --scm-project=demo --project=collection
        if (self.scm_org or self.scm_project) and self.project != "ansible-project":
            msg = (
                "The parameters 'scm-org' and 'scm-project' have no effect when"
                " project is not set to ansible-project."
            )
            self.output.warning(msg)

        # Validation for: ansible-creator init testorg.testname --project=ansible-project
        # --scm-org weather --scm-project demo
        if self.collection and self.project != "collection":
            msg = (
                "Collection name has no effect when project is set to ansible-project."
            )
            self.output.warning(msg)

        if self.collection:
            fqcn = self.collection.split(".", maxsplit=1)
            object.__setattr__(self, "namespace", fqcn[0])
            object.__setattr__(self, "collection_name", fqcn[-1])

        object.__setattr__(self, "init_path", expand_path(self.init_path))



"""Unit tests for ansible-creator init."""

from __future__ import annotations

import re
import sys
from textwrap import dedent


def test_run_help(cli):
    result = cli("ansible-creator --help")
    assert result.returncode == 0

    # temporary assertion fix until we write custom helper
    if sys.version_info < (3, 10):
        assert (
            dedent(
                """\
                usage: ansible-creator [-h] [--version] {init} ...

                Tool to scaffold Ansible Content. Get started by looking at the help text.

                optional arguments:
                  -h, --help  show this help message and exit
                  --version   Print ansible-creator version and exit.

                Commands:
                  {init}      The subcommand to invoke.
                    init      Initialize an Ansible Collection.
                """,
            )
            in result.stdout
        )
    else:
        assert (
            dedent(
                """\
                usage: ansible-creator [-h] [--version] {init} ...

                Tool to scaffold Ansible Content. Get started by looking at the help text.

                options:
                  -h, --help  show this help message and exit
                  --version   Print ansible-creator version and exit.

                Commands:
                  {init}      The subcommand to invoke.
                    init      Initialize an Ansible Collection.
                """,
            )
            in result.stdout
        )


def test_run_no_subcommand(cli):
    result = cli("ansible-creator")
    assert result.returncode != 0
    assert (
        dedent(
            """\
            usage: ansible-creator [-h] [--version] {init} ...
            ansible-creator: error: the following arguments are required: subcommand
            """,
        )
        in result.stderr
    )


def test_run_init_no_input(cli):
    result = cli("ansible-creator init")
    assert result.returncode != 0
    assert (
        "Error: The argument 'collection' is required when scaffolding a collection"
        in result.stderr
    )


def test_run_init_basic(cli, tmp_path):
    final_dest = f"{tmp_path}/collections/ansible_collections"
    cli(f"mkdir -p {final_dest}")

    result = cli(
        f"ansible-creator init testorg.testcol --init-path {final_dest}",
    )
    assert result.returncode == 0

    # check stdout
    assert (
        re.search("Note: collection testorg.testcol created at", result.stdout)
        is not None
    )

    # fail to override existing collection with force=false (default)
    result = cli(
        f"ansible-creator init testorg.testcol --init-path {final_dest}",
    )

    assert result.returncode != 0

    # this is required to handle random line breaks in CI, especially with macos runners
    mod_stderr = "".join([line.strip() for line in result.stderr.splitlines()])
    assert (
        re.search(
            rf"Error:\s*The\s*directory\s*{final_dest}/testorg/testcol\s*is\s*not\s*empty.",
            mod_stderr,
        )
        is not None
    )
    assert "You can use --force to re-initialize this directory." in result.stderr
    assert "However it will delete ALL existing contents in it." in result.stderr

    # override existing collection with force=true
    result = cli(f"ansible-creator init testorg.testcol --init-path {tmp_path} --force")
    assert result.returncode == 0
    assert (
        re.search("Warning: re-initializing existing directory", result.stdout)
        is not None
    )



"""Unit tests for ansible-creator init."""

from __future__ import annotations

import re

from filecmp import dircmp

import pytest

from ansible_creator.config import Config
from ansible_creator.exceptions import CreatorError
from ansible_creator.output import Output
from ansible_creator.subcommands.init import Init
from ansible_creator.utils import TermFeatures

from tests.defaults import FIXTURES_DIR


@pytest.fixture()
def cli_args(tmp_path, output: Output) -> dict:
    """Create an Init class object as fixture.

    :param tmp_path: App configuration object.
    """
    return {
        "creator_version": "0.0.1",
        "subcommand": "init",
        "collection": "testorg.testcol",
        "init_path": tmp_path / "testorg" / "testcol",
        "output": output,
    }


def test_run_success_for_collection(
    capsys,
    tmp_path,
    cli_args,
) -> None:
    """Test Init.run()."""
    # successfully create new collection

    cli_args["project"] = "collection"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # check stdout
    assert re.search("Note: collection testorg.testcol created", result) is not None

    # recursively assert files created
    dircmp(str(tmp_path), str(FIXTURES_DIR / "collection")).report_full_closure()
    captured = capsys.readouterr()
    assert re.search("Differing files|Only in", captured.out) is None, captured.out

    # fail to override existing collection with force=false (default)
    fail_msg = (
        f"The directory {tmp_path}/testorg/testcol is not empty."
        "\nYou can use --force to re-initialize this directory."
        "\nHowever it will delete ALL existing contents in it."
    )
    with pytest.raises(CreatorError, match=fail_msg):
        init.run()

    # override existing collection with force=true
    cli_args["force"] = True
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out
    assert (
        re.search("Warning: re-initializing existing directory", result) is not None
    ), result


def test_run_success_ansible_project(
    capsys,
    tmp_path,
    cli_args,
) -> None:
    """Test Init.run()."""
    # successfully create new ansible-project
    cli_args["collection"] = None
    cli_args["project"] = "ansible-project"
    cli_args["init_path"] = tmp_path / "new_project"
    cli_args["scm_org"] = "weather"
    cli_args["scm_project"] = "demo"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # check stdout
    assert re.search("Note: ansible project created", result) is not None

    # recursively assert files created
    dircmp(
        str(tmp_path / "new_project"),
        str(FIXTURES_DIR / "project" / "ansible_project"),
    ).report_full_closure()
    captured = capsys.readouterr()
    assert re.search("Differing files|Only in", captured.out) is None, captured.out

    # fail to override existing ansible-project directory with force=false (default)
    fail_msg = (
        f"The directory {tmp_path}/new_project is not empty."
        "\nYou can use --force to re-initialize this directory."
        "\nHowever it will delete ALL existing contents in it."
    )
    with pytest.raises(CreatorError, match=fail_msg):
        init.run()

    # override existing ansible-project directory with force=true
    cli_args["force"] = True
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out
    assert (
        re.search("Warning: re-initializing existing directory", result) is not None
    ), result


def test_run_success_collections_alt_dir(
    tmp_path,
    capsys,
    cli_args,
) -> None:
    """Test Init.run() when init_path ends with "collections" / "ansible_collections"""
    # successfully create new collection
    cli_args["project"] = "collection"
    cli_args["init_path"] = tmp_path / "collections" / "ansible_collections"
    final_path = cli_args["init_path"] / "testorg" / "testcol"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # this is required to handle random line breaks in CI, especially with macos runners
    mod_result = "".join([line.strip() for line in result.splitlines()])

    assert (
        re.search(
            rf"Note:\s*collection\s*testorg.testcol\s*created\s*at\s*{final_path}",
            mod_result,
        )
        is not None
    )


def test_error_1(
    tmp_path,
    cli_args,
) -> None:
    """Test Init.run()."""
    # Validation for: ansible-creator init --project=ansible-project
    cli_args["collection"] = None
    cli_args["project"] = "ansible-project"
    cli_args["init_path"] = tmp_path / "new_project"
    cli_args["scm_org"] = None
    cli_args["scm_project"] = None
    fail_msg = (
        "Parameters 'scm-org' and 'scm-project' are required when "
        "scaffolding an ansible-project."
    )
    with pytest.raises(CreatorError, match=fail_msg):
        init = Init(
            Config(**cli_args),
        )
        init.run()


def test_error_2(
    cli_args,
) -> None:
    """Test Init.run()."""
    # Validation for: ansible-creator init
    cli_args["collection"] = None
    cli_args["project"] = "collection"
    cli_args["init_path"] = None
    cli_args["scm_org"] = None
    cli_args["scm_project"] = None
    fail_msg = "The argument 'collection' is required when scaffolding a collection."
    with pytest.raises(CreatorError, match=fail_msg):
        init = Init(
            Config(**cli_args),
        )
        init.run()


def test_warning(
    capsys,
    tmp_path,
    cli_args,
) -> None:
    """Test Init.run()."""
    # Validation for: ansible-creator init testorg.testname --scm-org=weather
    # --scm-project=demo --project=collection
    cli_args["collection"] = "testorg.testname"
    cli_args["project"] = None
    cli_args["init_path"] = tmp_path / "testorg" / "testcol"
    cli_args["scm_org"] = "weather"
    cli_args["scm_project"] = "demo"
    init = Init(
        Config(**cli_args),
    )
    init.run()
    result = capsys.readouterr().out

    # this is required to handle random line breaks in CI, especially with macos runners
    mod_result = "".join([line.strip() for line in result.splitlines()])
    assert (
        re.search(
            rf"Warning:\s*The parameters\s*'scm-org'\s*and\s*'scm-project'"
            rf"\s*have\s*no\s*effect\s*when\s*project\s*is\s*not\s*set\s*to\s*ansible-project",
            mod_result,
        )
        is not None
    )



"""Definition of constants for this package."""

GLOBAL_TEMPLATE_VARS = {
    "DEV_CONTAINER_IMAGE": "ghcr.io/ansible/community-ansible-dev-tools:latest",
    "DEV_FILE_IMAGE": "ghcr.io/ansible/community-ansible-devspaces:latest",
}



"""Compat library for ansible-creator.

This contains compatibility definitions for older python
When we need to import a module differently depending on python versions, we do it
here.
"""

# ruff: noqa: F401

# pylint: disable=unused-import

import sys


if sys.version_info >= (3, 11):
    from importlib.resources.abc import Traversable
else:
    from importlib.abc import Traversable  # pylint: disable=W4904
